!
!     CalculiX - A 3-dimensional finite element program
!              Copyright (C) 1998-2015 Guido Dhondt
!
!     This program is free software; you can redistribute it and/or
!     modify it under the terms of the GNU General Public License as
!     published by the Free Software Foundation(version 2);
!     
!
!     This program is distributed in the hope that it will be useful,
!     but WITHOUT ANY WARRANTY; without even the implied warranty of 
!     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
!     GNU General Public License for more details.
!
!     You should have received a copy of the GNU General Public License
!     along with this program; if not, write to the Free Software
!     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
      subroutine mafillv0(nef,ipnei,neifa,neiel,vfa,xxn,area,
     &  auv6,adv,bv,vel,cosa,umfa,xlet,xle,gradvfa,xxi,body,volume,
     &  ielfa,ifabou,nbody,velo,sel,xxj,nactdohinv,a1,
     &  a2,flux,nefa,nefb,icyclic,c,ifatie,adv6)
!
!     sets up the velocity CFD-matrices for incompressible fluids
!     calculations with block structure
!
      implicit none
!
      integer i,nef,indexf,ipnei(*),j,ifa,iel,neifa(*),icyclic,
     &  neiel(*),iwall,ielfa(4,*),nefa,nefb,
     &  ipointer,ifabou(*),nbody,k,indexb,nactdohinv(*),
     &  ifatie(*)
!
      real*8 xflux,vfa(0:5,*),xxn(3,*),area(*),auv6(6,*),adv(*),
     &  vel(nef,0:5),cosa(*),umfa(*),xlet(*),xle(*),coef,gradvfa(3,3,*),
     &  xxi(3,*),body(0:3,*),volume(*),coef2,velo(nef,0:5),adv6(6,*),
     &  rhovel,constant,sel(3,*),xxj(3,*),a1,a2,
     &  flux(*),c(3,3),bv(3,nef)
!
      intent(in) nef,ipnei,neifa,neiel,vfa,xxn,area,
     &  vel,cosa,umfa,xlet,xle,gradvfa,xxi,
     &  body,volume,ielfa,ifabou,nbody,
     &  velo,xxj,nactdohinv,a1,
     &  a2,flux
!
      intent(inout) adv,auv6,bv,sel,adv6
!
      do i=nefa,nefb
         indexf=ipnei(i)
         do j=1,6
!
!              convection
!
            indexf=indexf+1
            ifa=neifa(indexf)
            iel=neiel(indexf)
            xflux=flux(indexf)
!
!              vfa sometimes undefined?
!
            if(xflux.ge.0.d0) then
!
!                 outflowing xflux
!
c               call add_sm_fl_as(auv,adv,jq,irow,i,i,
c     &              xflux,nzs)
               adv6(j,i)=adv6(j,i)+xflux
                  do k=1,3
                     bv(k,i)=bv(k,i)-(vfa(k,ifa)-vel(i,k))*xflux
                  enddo
            else
               if(iel.gt.0) then
!
!                    incoming flux from neighboring element
!
                  if((icyclic.eq.0).or.(ifatie(ifa).eq.0)) then
c                   call add_sm_fl_as(auv,adv,jq,irow,i,iel,xflux,
c     &                 nzs)
                     auv6(j,i)=auv6(j,i)+xflux
                     bv(1,i)=bv(1,i)-(vfa(1,ifa)-vel(iel,1))*xflux
                     bv(2,i)=bv(2,i)-(vfa(2,ifa)-vel(iel,2))*xflux
                     bv(3,i)=bv(3,i)-(vfa(3,ifa)-vel(iel,3))*xflux
                  else
                     bv(1,i)=bv(1,i)-vfa(1,ifa)*xflux
                     bv(2,i)=bv(2,i)-vfa(2,ifa)*xflux
                     bv(3,i)=bv(3,i)-vfa(3,ifa)*xflux
                  endif
               else
!
!                    incoming flux through boundary
!
                  if(ielfa(2,ifa).lt.0) then
                     indexb=-ielfa(2,ifa)
                     if(((ifabou(indexb+1).ne.0).and.
     &                    (ifabou(indexb+2).ne.0).and.
     &                    (ifabou(indexb+3).ne.0)).or.
     &                    (dabs(xflux).lt.1.d-10)) then
                        bv(1,i)=bv(1,i)-vfa(1,ifa)*xflux
                        bv(2,i)=bv(2,i)-vfa(2,ifa)*xflux
                        bv(3,i)=bv(3,i)-vfa(3,ifa)*xflux
                     else
                        write(*,*) '*ERROR in mafillv: not all'
                        write(*,*) '       components of an incoming'
                        write(*,*) '       flux through face ',j
                        write(*,*)'       of element ',nactdohinv(i),
     &                        ' are given',
     &vfa(1,ifa),vfa(2,ifa),vfa(3,ifa)
                     endif
                  else
                     write(*,*) '*ERROR in mafillv: not all'
                     write(*,*) '       components of an incoming'
                     write(*,*) '       flux through face ',j
                     write(*,*)'       of element ',nactdohinv(i),
     &                     ' are given',
     &vfa(1,ifa),vfa(2,ifa),vfa(3,ifa)
                  endif
               endif
            endif
!
!              diffusion
!
            if(iel.ne.0) then
!
!                 neighboring element
!
               coef=umfa(ifa)*area(ifa)/xlet(indexf)
c               call add_sm_fl_as(auv,adv,jq,irow,i,i,
c     &              coef,nzs)
               adv6(j,i)=adv6(j,i)+coef
               if((icyclic.eq.0).or.(ifatie(ifa).eq.0)) then
c                  call add_sm_fl_as(auv,adv,jq,irow,i,iel,
c     &              -coef,nzs)
                  auv6(j,i)=auv6(j,i)-coef
               elseif(ifatie(ifa).gt.0) then
!
!                 for cyclic symmetry the term is retarded, since
!                 otherwise the x-, y- and z- components are linked
!                 (i.e. the x-, y- and z- momentum equations cannot
!                  be solved separately any more)
!
                  do k=1,3
                     bv(k,i)=bv(k,i)+
     &                (c(k,1)*vel(iel,1)+c(k,2)*vel(iel,2)+
     &                 c(k,3)*vel(iel,3))*coef
                  enddo
               else
                  do k=1,3
                     bv(k,i)=bv(k,i)+
     &                (c(1,k)*vel(iel,1)+c(2,k)*vel(iel,2)+
     &                 c(3,k)*vel(iel,3))*coef
                  enddo
               endif
!
!                 correction for non-orthogonal grid
!
               bv(1,i)=bv(1,i)+umfa(ifa)*area(ifa)*
     &              (gradvfa(1,1,ifa)*(xxn(1,indexf)-xxj(1,indexf))+
     &              gradvfa(1,2,ifa)*(xxn(2,indexf)-xxj(2,indexf))+
     &              gradvfa(1,3,ifa)*(xxn(3,indexf)-xxj(3,indexf)))
               bv(2,i)=bv(2,i)+umfa(ifa)*area(ifa)*
     &              (gradvfa(2,1,ifa)*(xxn(1,indexf)-xxj(1,indexf))+
     &              gradvfa(2,2,ifa)*(xxn(2,indexf)-xxj(2,indexf))+
     &              gradvfa(2,3,ifa)*(xxn(3,indexf)-xxj(3,indexf)))
               bv(3,i)=bv(3,i)+umfa(ifa)*area(ifa)*
     &              (gradvfa(3,1,ifa)*(xxn(1,indexf)-xxj(1,indexf))+
     &              gradvfa(3,2,ifa)*(xxn(2,indexf)-xxj(2,indexf))+
     &              gradvfa(3,3,ifa)*(xxn(3,indexf)-xxj(3,indexf)))
            else
!
!                 boundary; check whether wall (specified by user),
!                           outlet (no velocity boundary conditions or
!                           none of those
!
               iwall=0
               ipointer=abs(ielfa(2,ifa))
               if(ipointer.gt.0) then
                  iwall=ifabou(ipointer+5)
               endif
               if(iwall.ne.1) then
!
!                    external face, but no wall
!
                  if((ifabou(ipointer+1).ne.0).or.
     &                 (ifabou(ipointer+2).ne.0).or.
     &                 (ifabou(ipointer+3).ne.0)) then
!
!                       no outlet: face velocity fixed
!
                     coef=umfa(ifa)*area(ifa)/xle(indexf)
c                     call add_sm_fl_as(auv,adv,jq,irow,i,i,
c     &                    coef,nzs)
                     adv6(j,i)=adv6(j,i)+coef
                     bv(1,i)=bv(1,i)+coef*vfa(1,ifa)
                     bv(2,i)=bv(2,i)+coef*vfa(2,ifa)
                     bv(3,i)=bv(3,i)+coef*vfa(3,ifa)
                  else
!
!                       outlet: no diffusion
!
                  endif
!
!                    correction for non-orthogonal grid
!
                  bv(1,i)=bv(1,i)+umfa(ifa)*area(ifa)*
     &                 (gradvfa(1,1,ifa)*(xxn(1,indexf)-xxi(1,indexf))+
     &                 gradvfa(1,2,ifa)*(xxn(2,indexf)-xxi(2,indexf))+
     &                 gradvfa(1,3,ifa)*(xxn(3,indexf)-xxi(3,indexf)))
                  bv(2,i)=bv(2,i)+umfa(ifa)*area(ifa)*
     &                 (gradvfa(2,1,ifa)*(xxn(1,indexf)-xxi(1,indexf))+
     &                 gradvfa(2,2,ifa)*(xxn(2,indexf)-xxi(2,indexf))+
     &                 gradvfa(2,3,ifa)*(xxn(3,indexf)-xxi(3,indexf)))
                  bv(3,i)=bv(3,i)+umfa(ifa)*area(ifa)*
     &                 (gradvfa(3,1,ifa)*(xxn(1,indexf)-xxi(1,indexf))+
     &                 gradvfa(3,2,ifa)*(xxn(2,indexf)-xxi(2,indexf))+
     &                 gradvfa(3,3,ifa)*(xxn(3,indexf)-xxi(3,indexf)))
               else
!     
!                    wall
!     
                  coef=umfa(ifa)*area(ifa)/(xle(indexf)*cosa(indexf))
c                  call add_sm_fl_as(auv,adv,jq,irow,i,i,coef,
c     &                 nzs)
                  adv6(j,i)=adv6(j,i)+coef
!
!                    correction for non-orthogonal grid and nonzero
!                    wall velocity
!
                  coef2=((vel(i,1)-vfa(1,ifa))*xxn(1,indexf)+
     &                 (vel(i,2)-vfa(2,ifa))*xxn(2,indexf)+
     &                 (vel(i,3)-vfa(3,ifa))*xxn(3,indexf))*coef
                  bv(1,i)=bv(1,i)+coef*vfa(1,ifa)+
     &                 coef2*xxn(1,indexf)
                  bv(2,i)=bv(2,i)+coef*vfa(2,ifa)+
     &                 coef2*xxn(2,indexf)
                  bv(3,i)=bv(3,i)+coef*vfa(3,ifa)+
     &                 coef2*xxn(3,indexf)
               endif
            endif
!     
!     pressure
!     
         enddo
!     
!           body force
!     
         rhovel=vel(i,5)*volume(i)
!
         if(nbody.gt.0) then
            bv(1,i)=bv(1,i)+rhovel*body(1,i)
            bv(2,i)=bv(2,i)+rhovel*body(2,i)
            bv(3,i)=bv(3,i)+rhovel*body(3,i)
         endif
!
!           transient term
!
         constant=rhovel/3.d0
         bv(1,i)=bv(1,i)-a2*velo(i,1)*constant
         bv(2,i)=bv(2,i)-a2*velo(i,2)*constant
         bv(3,i)=bv(3,i)-a2*velo(i,3)*constant
         constant=a1*constant
c         call add_sm_fl(auv,adv,jq,irow,i,i,constant,nzs)
         adv(i)=adv(i)+constant
!
!           copying b into sel (rhs without pressure)
!
         do j=1,3
            sel(j,i)=bv(j,i)
         enddo
!
!           pressure contribution to b
!
         indexf=ipnei(i)
         do j=1,6
            indexf=indexf+1
            ifa=neifa(indexf)
            bv(1,i)=bv(1,i)
     &           -vfa(4,ifa)*xxn(1,indexf)*area(ifa)
            bv(2,i)=bv(2,i)
     &           -vfa(4,ifa)*xxn(2,indexf)*area(ifa)
            bv(3,i)=bv(3,i)
     &           -vfa(4,ifa)*xxn(3,indexf)*area(ifa)
         enddo
!            
      enddo
!     
      return
      end
